"""Python implementation of sigrok tools"""
from enum import Enum
import sys
if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

__version__ = "0.0.3"

class OutputType(Enum):
    SRD_OUTPUT_ANN = 0
    SRD_OUTPUT_PYTHON = 1
    SRD_OUTPUT_BINARY = 2
    SRD_OUTPUT_LOGIC = 3
    SRD_OUTPUT_META = 4

OUTPUT_ANN = OutputType.SRD_OUTPUT_ANN
OUTPUT_PYTHON = OutputType.SRD_OUTPUT_PYTHON
OUTPUT_BINARY = OutputType.SRD_OUTPUT_BINARY
OUTPUT_LOGIC = OutputType.SRD_OUTPUT_LOGIC
OUTPUT_META = OutputType.SRD_OUTPUT_META

class Decoder:
    # __init__() won't get called by subclasses

    def register(self, output_type, meta=None):
        # print("register", output_type, meta)
        return output_type

    def add_callback(self, output_type, fun):
        if not hasattr(self, "callbacks"):
            self.callbacks = {}

        if output_type not in self.callbacks:
            self.callbacks[output_type] = set()

        self.callbacks[output_type].add(fun)

    def wait(self, conds=None):
        assert(hasattr(self, "input"))
        if isinstance(conds, dict):
            conds = [conds]
        if self.one_to_one:
            data_conds = conds
        else:
            data_conds = []
            for cond in conds:
                data_cond = {}
                for k in cond:
                    data_cond[self.decoder_channel_to_data_channel[k]] = cond[k]
                data_conds.append(data_cond)
        raw_data = self.input.wait(data_conds)
        data = [None] * (len(self.channels) + len(getattr(self, "optional_channels", [])))
        for decoder_channel in self.decoder_channel_to_data_channel:
            data_channel = self.decoder_channel_to_data_channel[decoder_channel]
            data[decoder_channel] = raw_data[data_channel]
        return tuple(data)

    def put(self, startsample, endsample, output_id, data):
        # print(startsample, endsample, output_id, data)
        if not output_id in self.callbacks:
            return
        for cb in self.callbacks[output_id]:
            cb(startsample, endsample, data)

    def set_channelnum(self, channelname, channelnum):
        if not hasattr(self, "decoder_channel_to_data_channel"):
            self.decoder_channel_to_data_channel = {}
            self.one_to_one = True
        optional = tuple()
        if hasattr(self, "optional_channels"):
            optional = self.optional_channels
        for i, c in enumerate(self.channels + optional):
            if c["id"] == channelname:
                self.decoder_channel_to_data_channel[i] = channelnum
                self.one_to_one = self.one_to_one and i == channelnum
                break

    def has_channel(self, decoder_channel):
        return decoder_channel in self.decoder_channel_to_data_channel

    @property
    def samplenum(self):
        return self.input.samplenum

    @property
    def matched(self):
        return self.input.matched

def get_decoder(decoder_id):
    discovered_plugins = entry_points(name=decoder_id, group='pysigrok.decoders')
    if len(discovered_plugins) == 1:
        return discovered_plugins[0].load()
    raise RuntimeError("Decoder id ambiguous:" + ",".join([p.name for p in discovered_plugins]))



import sys
if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

discovered_plugins = entry_points(group='pysigrok.decoders')

def main():
    print("hello world")
    for p in discovered_plugins:
        print(p)
        loaded = p.load()
        print(loaded)
        print(loaded.name, loaded.longname, loaded.desc)



import zipfile
import configparser

class SrZip:
    def __init__(self, filename, initial_state=None):
        self.zip = zipfile.ZipFile(filename)
        # self.zip.printdir()
        self.metadata = configparser.ConfigParser()
        self.version = int(self.zip.read("version").decode("ascii"))
        self.metadata.read_string(self.zip.read("metadata").decode("ascii"))
        self.samplenum = -1
        self.matched = None
        # print(self.version)
        # for s in self.metadata.sections():
        #     print(s)
        #     for o in self.metadata.options(s):
        #         print(" ", o, self.metadata.get(s, o))

        self.last_sample = 0
        for channel in initial_state:
            self.last_sample |= initial_state[channel] << channel
        self.unitsize = int(self.metadata.get("device 1", "unitsize"))
        assert(self.unitsize == 1)

        if self.version == 1:
            self.data = self.zip.read("logic-1")
        else:
            self.data = None
            self._file_start = -1
            self._file_index = 1

    def wait(self, conds=None):
        self.matched = [False]
        while not any(self.matched):
            self.matched = [True] * len(conds)
            self.samplenum += 1
            if self.version == 1:
                if self.samplenum * self.unitsize >= len(self.data):
                    raise EOFError()
                sample = self.data[self.samplenum]
            elif self.version == 2:
                file_samplenum = self.samplenum - self._file_start
                if self.data is None or file_samplenum * self.unitsize >= len(self.data):
                    self._file_start = self.samplenum
                    try:
                        self.data = self.data = self.zip.read(f"logic-1-{self._file_index:d}")
                    except KeyError:
                        raise EOFError()
                    file_samplenum = 0
                    self._file_index += 1
                sample = self.data[file_samplenum]

            for i, cond in enumerate(conds):
                for channel in cond:
                    state = cond[channel]
                    mask = 1 << channel
                    last_value = self.last_sample & mask
                    value = sample & mask
                    if ((state == "l" and value != 0) or
                        (state == "h" and value == 0) or
                        (state == "r" and not (last_value == 0 and value != 0)) or
                        (state == "f" and not (last_value != 0 and value == 0)) or
                        (state == "e" and last_value == value) or
                        (state == "s" and last_value != value)):
                            self.matched[i] = False
                            break
            self.last_sample = sample

        bits = []
        for b in range(self.unitsize * 8):
            bits.append((sample >> b) & 0x1)

        return tuple(bits)




import click

import functools

from . import srzip
from . import *

def output_python(decoder, outfilter, outfile, startsample, endsample, data):
    outfile.write(f"{startsample}-{endsample} {decoder.id}: {data}\n".encode("utf-8"))

def output_annotation(decoder, outfilter, outfile, startsample, endsample, data):
    annotation = decoder.annotations[data[0]]
    if outfilter is None or annotation[0] == outfilter:
        data = " ".join([repr(x) for x in data[1]]).replace("\'", "\"")
        outfile.write(f"{startsample}-{endsample} {decoder.id}: {annotation[0]}: {data}\n".encode("utf-8"))

def output_binary(decoder, outfilter, outfile, startsample, endsample, data):
    track = decoder.binary[data[0]]
    if track[0] == outfilter:
        data = " ".join([f"{x:02x}" for x in data[1]])
        outfile.write(f"{startsample}-{endsample} {decoder.id}: {data}\n".encode("utf-8"))

OUTPUT_TYPES = {
    "python": (OUTPUT_PYTHON, output_python),
    "annotation": (OUTPUT_ANN, output_annotation),
    "binary": (OUTPUT_BINARY, output_binary)
}

@click.command()
@click.option('-P', "--protocol-decoder")
@click.option("-p", "--pin-mapping", multiple=True, help="channelname=channelnum")
@click.option("-o", "--channel-option", multiple=True, help="channeloption=value")
@click.option("-N", "--channel-initial-value", multiple=True, help="channelname=initial-pin-value")
@click.option("-i", "--input-file")
@click.option("-O", "--output-format")
@click.option("-f", "--output-file")
def main(protocol_decoder, pin_mapping, channel_option, channel_initial_value, input_file, output_format, output_file):
    parsed_mapping = {}
    for mapping in pin_mapping:
        decoder_id, channelnum = mapping.split("=")
        channelnum = int(channelnum)
        parsed_mapping[decoder_id] = channelnum

    initial_values = {}
    for v in channel_initial_value:
        decoder_id, value = v.split("=")
        value = int(value)
        initial_values[parsed_mapping[decoder_id]] = value


    data = srzip.SrZip(input_file, initial_values)
    decoder_class = get_decoder(protocol_decoder)
    decoder = decoder_class()
    decoder.options = {}
    if hasattr(decoder_class, "options"):
        for o in decoder_class.options:
            decoder.options[o["id"]] = o["default"]
    decoder.input = data
    for decoder_id in parsed_mapping:
        channelnum = parsed_mapping[decoder_id]
        decoder.set_channelnum(decoder_id, channelnum)

    with open(output_file, "wb") as f:
        if output_format.count(":") == 1:
            decoder_id, output_name = output_format.split(":")
            output_filter = None
        else:
            decoder_id, output_name, output_filter = output_format.split(":", maxsplit=2)
        output_type, output_fun = OUTPUT_TYPES[output_name]
        decoder.add_callback(output_type, functools.partial(output_fun, decoder, output_filter, f))
        decoder.start()
        decoder.reset()
        try:
            decoder.decode()
        except EOFError:
            pass

